/*
 * #%L
 * SchemaCrawler Maven Plugin
 * %%
 * Copyright (C) 2011 SchemaCrawler
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation, either version 3 of the 
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Lesser Public License for more details.
 * 
 * You should have received a copy of the GNU General Lesser Public 
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/lgpl-3.0.html>.
 * #L%
 */
package schemacrawler.tools.integration.maven;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;

/**
 * Display help information on schemacrawler-maven-plugin.<br/> Call <pre>  mvn schemacrawler:help -Ddetail=true -Dgoal=&lt;goal-name&gt;</pre> to display parameter details.
 *
 * @version generated on Fri Nov 11 00:48:46 PST 2011
 * @author org.apache.maven.tools.plugin.generator.PluginHelpGenerator (version 2.9)
 * @goal help
 * @requiresProject false
 * @threadSafe
 */
@SuppressWarnings( "all" )
public class HelpMojo
    extends AbstractMojo
{
    /**
     * If <code>true</code>, display all settable properties for each goal.
     * 
     * @parameter expression="${detail}" default-value="false"
     */
    private boolean detail;

    /**
     * The name of the goal for which to show help. If unspecified, all goals will be displayed.
     * 
     * @parameter expression="${goal}"
     */
    private java.lang.String goal;

    /**
     * The maximum length of a display line, should be positive.
     * 
     * @parameter expression="${lineLength}" default-value="80"
     */
    private int lineLength;

    /**
     * The number of spaces per indentation level, should be positive.
     * 
     * @parameter expression="${indentSize}" default-value="2"
     */
    private int indentSize;


    /** {@inheritDoc} */
    public void execute()
        throws MojoExecutionException
    {
        if ( lineLength <= 0 )
        {
            getLog().warn( "The parameter 'lineLength' should be positive, using '80' as default." );
            lineLength = 80;
        }
        if ( indentSize <= 0 )
        {
            getLog().warn( "The parameter 'indentSize' should be positive, using '2' as default." );
            indentSize = 2;
        }

        StringBuffer sb = new StringBuffer();

        append( sb, "net.sourceforge.schemacrawler:schemacrawler-maven-plugin:1.0", 0 );
        append( sb, "", 0 );

        append( sb, "SchemaCrawler Maven Plugin", 0 );
        append( sb, "SchemaCrawler is an open-source Java API that makes working with database metadata as easy as working with plain old Java objects. SchemaCrawler is also a database schema discovery and comprehension, and schema documentation tool. You can search for database schema objects using regular expressions, and output the schema and data in a readable text format. The output is designed to be diff-ed against other database schemas.", 1 );
        append( sb, "", 0 );

        if ( goal == null || goal.length() <= 0 )
        {
            append( sb, "This plugin has 2 goals:", 0 );
            append( sb, "", 0 );
        }

        if ( goal == null || goal.length() <= 0 || "help".equals( goal ) )
        {
            append( sb, "schemacrawler:help", 0 );
            append( sb, "Display help information on schemacrawler-maven-plugin.\nCall\n\u00a0\u00a0mvn\u00a0schemacrawler:help\u00a0-Ddetail=true\u00a0-Dgoal=<goal-name>\nto display parameter details.", 1 );
            append( sb, "", 0 );
            if ( detail )
            {
                append( sb, "Available parameters:", 1 );
                append( sb, "", 0 );

                append( sb, "detail (Default: false)", 2 );
                append( sb, "If true, display all settable properties for each goal.", 3 );
                append( sb, "Expression: ${detail}", 3 );
                append( sb, "", 0 );

                append( sb, "goal", 2 );
                append( sb, "The name of the goal for which to show help. If unspecified, all goals will be displayed.", 3 );
                append( sb, "Expression: ${goal}", 3 );
                append( sb, "", 0 );

                append( sb, "indentSize (Default: 2)", 2 );
                append( sb, "The number of spaces per indentation level, should be positive.", 3 );
                append( sb, "Expression: ${indentSize}", 3 );
                append( sb, "", 0 );

                append( sb, "lineLength (Default: 80)", 2 );
                append( sb, "The maximum length of a display line, should be positive.", 3 );
                append( sb, "Expression: ${lineLength}", 3 );
                append( sb, "", 0 );
            }
        }

        if ( goal == null || goal.length() <= 0 || "schemacrawler".equals( goal ) )
        {
            append( sb, "schemacrawler:schemacrawler", 0 );
            append( sb, "Generates a SchemaCrawler report of the database.", 1 );
            append( sb, "", 0 );
            if ( detail )
            {
                append( sb, "Available parameters:", 1 );
                append( sb, "", 0 );

                append( sb, "command", 2 );
                append( sb, "Command.", 3 );
                append( sb, "Required: Yes", 3 );
                append( sb, "Expression: ${command}", 3 );
                append( sb, "", 0 );

                append( sb, "config (Default: config.properties)", 2 );
                append( sb, "Config file.", 3 );
                append( sb, "Required: Yes", 3 );
                append( sb, "Expression: ${config}", 3 );
                append( sb, "", 0 );

                append( sb, "configOverride (Default: schemacrawler.config.override.properties)", 2 );
                append( sb, "Config override file.", 3 );
                append( sb, "Expression: ${config-override}", 3 );
                append( sb, "", 0 );

                append( sb, "driver", 2 );
                append( sb, "JDBC driver class name.", 3 );
                append( sb, "Required: Yes", 3 );
                append( sb, "Expression: ${driver}", 3 );
                append( sb, "", 0 );

                append( sb, "excludecolumns", 2 );
                append( sb, "Regular expression to match fully qualified column names, in the form \'CATALOGNAME.SCHEMANAME.TABLENAME.COLUMNNAME\' - for example, .*\\.STREET|.*\\.PRICE matches columns named STREET or PRICE in any table Columns that match the pattern are not displayed", 3 );
                append( sb, "Expression: ${excludecolumns}", 3 );
                append( sb, "", 0 );

                append( sb, "excludeinout", 2 );
                append( sb, "Regular expression to match fully qualified parameter names. Parameters that match the pattern are not displayed", 3 );
                append( sb, "Expression: ${excludeinout}", 3 );
                append( sb, "", 0 );

                append( sb, "infolevel", 2 );
                append( sb, "The info level determines the amount of database metadata retrieved, and also determines the time taken to crawl the schema.", 3 );
                append( sb, "Required: Yes", 3 );
                append( sb, "Expression: ${infolevel}", 3 );
                append( sb, "", 0 );

                append( sb, "password", 2 );
                append( sb, "Database connection user password.", 3 );
                append( sb, "Expression: ${password}", 3 );
                append( sb, "", 0 );

                append( sb, "procedures", 2 );
                append( sb, "Regular expression to match fully qualified procedure names, in the form \'CATALOGNAME.SCHEMANAME.PROCEDURENAME\' - for example, .*\\.C.*|.*\\.P.* matches any procedures whose names start with C or P Procedures that do not match the pattern are not displayed", 3 );
                append( sb, "Expression: ${procedures}", 3 );
                append( sb, "", 0 );

                append( sb, "schemas", 2 );
                append( sb, "(no description available)", 3 );
                append( sb, "Expression: ${schemas}", 3 );
                append( sb, "", 0 );

                append( sb, "sortcolumns (Default: false)", 2 );
                append( sb, "Sort columns in a table alphabetically.", 3 );
                append( sb, "Expression: ${sortcolumns}", 3 );
                append( sb, "", 0 );

                append( sb, "sortinout (Default: false)", 2 );
                append( sb, "Sort parameters in a stored procedure alphabetically.", 3 );
                append( sb, "Expression: ${sortinout}", 3 );
                append( sb, "", 0 );

                append( sb, "sorttables (Default: true)", 2 );
                append( sb, "Sort tables alphabetically.", 3 );
                append( sb, "Expression: ${sorttables}", 3 );
                append( sb, "", 0 );

                append( sb, "table_types", 2 );
                append( sb, "Comma-separated list of table types of TABLE,VIEW,SYSTEM_TABLE,GLOBAL_TEMPORARY,LOCAL_TEMPORARY,ALIAS", 3 );
                append( sb, "Expression: ${table_types}", 3 );
                append( sb, "", 0 );

                append( sb, "tables", 2 );
                append( sb, "Regular expression to match fully qualified table names, in the form \'CATALOGNAME.SCHEMANAME.TABLENAME\' - for example, .*\\.C.*|.*\\.P.* Tables that do not match the pattern are not displayed.", 3 );
                append( sb, "Expression: ${tables}", 3 );
                append( sb, "", 0 );

                append( sb, "url", 2 );
                append( sb, "Database connection string.", 3 );
                append( sb, "Required: Yes", 3 );
                append( sb, "Expression: ${url}", 3 );
                append( sb, "", 0 );

                append( sb, "user", 2 );
                append( sb, "Database connection user name.", 3 );
                append( sb, "Required: Yes", 3 );
                append( sb, "Expression: ${user}", 3 );
                append( sb, "", 0 );
            }
        }

        if ( getLog().isInfoEnabled() )
        {
            getLog().info( sb.toString() );
        }
    }

    /**
     * <p>Repeat a String <code>n</code> times to form a new string.</p>
     *
     * @param str String to repeat
     * @param repeat number of times to repeat str
     * @return String with repeated String
     * @throws NegativeArraySizeException if <code>repeat < 0</code>
     * @throws NullPointerException if str is <code>null</code>
     */
    private static String repeat( String str, int repeat )
    {
        StringBuffer buffer = new StringBuffer( repeat * str.length() );

        for ( int i = 0; i < repeat; i++ )
        {
            buffer.append( str );
        }

        return buffer.toString();
    }

    /** 
     * Append a description to the buffer by respecting the indentSize and lineLength parameters.
     * <b>Note</b>: The last character is always a new line.
     * 
     * @param sb The buffer to append the description, not <code>null</code>.
     * @param description The description, not <code>null</code>.
     * @param indent The base indentation level of each line, must not be negative.
     */
    private void append( StringBuffer sb, String description, int indent )
    {
        for ( Iterator it = toLines( description, indent, indentSize, lineLength ).iterator(); it.hasNext(); )
        {
            sb.append( it.next().toString() ).append( '\n' );
        }
    }

    /** 
     * Splits the specified text into lines of convenient display length.
     * 
     * @param text The text to split into lines, must not be <code>null</code>.
     * @param indent The base indentation level of each line, must not be negative.
     * @param indentSize The size of each indentation, must not be negative.
     * @param lineLength The length of the line, must not be negative.
     * @return The sequence of display lines, never <code>null</code>.
     * @throws NegativeArraySizeException if <code>indent < 0</code>
     */
    private static List toLines( String text, int indent, int indentSize, int lineLength )
    {
        List<String> lines = new ArrayList<String>();

        String ind = repeat( "\t", indent );
        String[] plainLines = text.split( "(\r\n)|(\r)|(\n)" );
        for ( int i = 0; i < plainLines.length; i++ )
        {
            toLines( lines, ind + plainLines[i], indentSize, lineLength );
        }

        return lines;
    }

    /** 
     * Adds the specified line to the output sequence, performing line wrapping if necessary.
     * 
     * @param lines The sequence of display lines, must not be <code>null</code>.
     * @param line The line to add, must not be <code>null</code>.
     * @param indentSize The size of each indentation, must not be negative.
     * @param lineLength The length of the line, must not be negative.
     */
    private static void toLines( List<String> lines, String line, int indentSize, int lineLength )
    {
        int lineIndent = getIndentLevel( line );
        StringBuffer buf = new StringBuffer( 256 );
        String[] tokens = line.split( " +" );
        for ( int i = 0; i < tokens.length; i++ )
        {
            String token = tokens[i];
            if ( i > 0 )
            {
                if ( buf.length() + token.length() >= lineLength )
                {
                    lines.add( buf.toString() );
                    buf.setLength( 0 );
                    buf.append( repeat( " ", lineIndent * indentSize ) );
                }
                else
                {
                    buf.append( ' ' );
                }
            }
            for ( int j = 0; j < token.length(); j++ )
            {
                char c = token.charAt( j );
                if ( c == '\t' )
                {
                    buf.append( repeat( " ", indentSize - buf.length() % indentSize ) );
                }
                else if ( c == '\u00A0' )
                {
                    buf.append( ' ' );
                }
                else
                {
                    buf.append( c );
                }
            }
        }
        lines.add( buf.toString() );
    }

    /** 
     * Gets the indentation level of the specified line.
     * 
     * @param line The line whose indentation level should be retrieved, must not be <code>null</code>.
     * @return The indentation level of the line.
     */
    private static int getIndentLevel( String line )
    {
        int level = 0;
        for ( int i = 0; i < line.length() && line.charAt( i ) == '\t'; i++ )
        {
            level++;
        }
        for ( int i = level + 1; i <= level + 4 && i < line.length(); i++ )
        {
            if ( line.charAt( i ) == '\t' )
            {
                level++;
                break;
            }
        }
        return level;
    }
}
